# -*- test-case-name: mimic.test.test_auth -*-
"""
Defines get current customer
"""

import json

from mimic.rest.mimicapp import MimicApp
from mimic.canned_responses import fastly


class FastlyApi(object):

    """
    Rest endpoints for mocked Fastly api.
    """

    app = MimicApp()

    def __init__(self, core):
        """
        :param MimicCore core: The core to which this FastlyApi will be
            communicating.
        """
        self.core = core
        self.services = {}
        self.fastly_response = fastly.FastlyResponse()

    @app.route('/', methods=['GET'])
    def get_health(self, request):
        """
        Returns response with 200 OK.
        """
        response = self.fastly_response.get_health()
        return json.dumps(response)

    @app.route('/current_customer', methods=['GET'])
    def get_current_customer(self, request):
        """
        Returns response with current customer details.

        https://docs.fastly.com/api/account#customer_1
        """
        response = self.fastly_response.get_current_customer()
        return json.dumps(response)

    @app.route('/service', methods=['POST'])
    def create_service(self, request):
        """
        Returns POST Service.

        https://docs.fastly.com/api/config#service_5
        """
        url_data = request.args.items()
        response = self.fastly_response.create_service(url_data)
        return json.dumps(response)

    @app.route('/service/<string:service_id>/version', methods=['POST'])
    def create_version(self, request, service_id):
        """
        Returns POST Service.

        https://docs.fastly.com/api/config#version_2
        """
        response = self.fastly_response.create_version(service_id)
        return json.dumps(response)

    @app.route('/service/search', methods=['GET'])
    def get_service_by_name(self, request):
        """
        Returns response with current customer details.

        https://docs.fastly.com/api/config#service_3
        """
        url_data = request.args.items()
        data = dict((key, value) for key, value in url_data)
        service_name = data['name'][0]

        response = self.fastly_response.get_service_by_name(service_name)
        return json.dumps(response)

    @app.route(
        '/service/<string:service_id>/version/<string:version_id>/domain',
        methods=['POST'])
    def create_domain(self, request, service_id, version_id):
        """
        Returns Create Domain Response.

        https://docs.fastly.com/api/config#domain_4
        """
        url_data = request.args.items()
        response = self.fastly_response.create_domain(url_data,
                                                      service_id, version_id)
        return json.dumps(response)

    @app.route(
        '/service/<string:service_id>/version/<string:version_id>/domain/'
        'check_all',
        methods=['GET'])
    def check_domains(self, request, service_id, version_id):
        """
        Returns Check Domain.

        https://docs.fastly.com/api/config#domain_3
        """
        response = self.fastly_response.check_domains(service_id, version_id)
        return json.dumps(response)

    @app.route(
        '/service/<string:service_id>/version/<string:version_id>/backend',
        methods=['POST'])
    def create_backend(self, request, service_id, version_id):
        """
        Returns Create Backend Response.

        https://docs.fastly.com/api/config#backend_2
        """
        url_data = request.args.items()
        response = self.fastly_response.create_backend(url_data,
                                                       service_id, version_id)
        return json.dumps(response)

    @app.route(
        '/service/<string:service_id>/version/<string:version_id>/condition',
        methods=['POST'])
    def create_condition(self, request, service_id, version_id):
        """
        Returns Create Condition Response.

        https://docs.fastly.com/api/config#condition_3
        """
        url_data = request.args.items()
        response = self.fastly_response.create_condition(url_data,
                                                         service_id, version_id)
        return json.dumps(response)

    @app.route(
        '/service/<string:service_id>/version/<string:version_id>/cache_settings',
        methods=['POST'])
    def create_cache_settings(self, request, service_id, version_id):
        """
        Returns Create Cache Settings Response.

        https://docs.fastly.com/api/config#cache_settings_3
        """
        url_data = request.args.items()
        response = self.fastly_response.create_cache_settings(url_data,
                                                              service_id, version_id)
        return json.dumps(response)

    @app.route(
        '/service/<string:service_id>/version/<string:version_id>/response_object',
        methods=['POST'])
    def create_response_object(self, request, service_id, version_id):
        """
        Returns Create Cache Settings Response.

        https://docs.fastly.com/api/config#response_object_3
        """
        url_data = request.args.items()
        response = self.fastly_response.create_response_object(url_data,
                                                               service_id, version_id)
        return json.dumps(response)

    @app.route(
        '/service/<string:service_id>/version/<string:version_id>/settings',
        methods=['PUT'])
    def create_settings(self, request, service_id, version_id):
        """
        Returns Settings Response.

        https://docs.fastly.com/api/config#settings_2
        """
        url_data = request.args.items()
        response = self.fastly_response.create_settings(url_data,
                                                        service_id, version_id)
        return json.dumps(response)

    @app.route('/service/<string:service_id>/version', methods=['GET'])
    def list_versions(self, request, service_id):
        """
        Returns List of Service versions.

        https://docs.fastly.com/api/config#version_3
        """
        response = self.fastly_response.list_versions(service_id)
        return json.dumps(response)

    @app.route('/service/<string:service_id>/version/<string:version_number>/'
               'activate', methods=['PUT'])
    def activate_version(self, request, service_id, version_number):
        """
        Returns Activate Service versions.

        https://docs.fastly.com/api/config#version_5
        """
        response = self.fastly_response.activate_version(service_id,
                                                         version_number)
        return json.dumps(response)

    @app.route('/service/<string:service_id>/version/<string:version_number>/'
               'deactivate', methods=['PUT'])
    def deactivate_version(self, request, service_id, version_number):
        """
        Returns Activate Service versions.

        https://docs.fastly.com/api/config#version_6
        """
        response = self.fastly_response.deactivate_version(service_id,
                                                           version_number)
        return json.dumps(response)

    @app.route('/service/<string:service_id>', methods=['DELETE'])
    def delete_service(self, request, service_id):
        """
        Returns DELETE Service.

        https://docs.fastly.com/api/config#service_6
        """
        response = self.fastly_response.delete_service(service_id)
        return json.dumps(response)

    @app.route('/service/<string:service_id>/details', methods=['GET'])
    def get_service_details(self, request, service_id):
        """
        Returns Service details.

        https://docs.fastly.com/api/config#service_2
        """
        response = self.fastly_response.get_service_details(service_id)
        return json.dumps(response)



"""
Model objects for the CLB mimic.  Please see the `Rackspace Cloud Load
Balancer API docs
<http://docs.rackspace.com/loadbalancers/api/v1.0/clb-devguide/content/API_Operations.html>`
 for more information.
"""
from copy import deepcopy
from random import randrange

import attr

from characteristic import attributes, Attribute

from six import string_types

from toolz.dicttoolz import dissoc

from twisted.internet.interfaces import IReactorTime
from twisted.python import log

from mimic.canned_responses.loadbalancer import load_balancer_example
from mimic.model.clb_errors import (
    considered_immutable_error,
    invalid_json_schema,
    loadbalancer_not_found,
    lb_deleted_xml,
    node_not_found,
    not_found_xml,
    updating_node_validation_error
)
from mimic.util.helper import (EMPTY_RESPONSE,
                               invalid_resource,
                               not_found_response,
                               seconds_to_timestamp,
                               set_resource_status)


def _one_of_validator(*items):
    """
    Return an :mod:`attr` validator which raises a :class:`TypeError`
    if the value is not equivalent to one of the provided items.

    :param items: Items to compare against
    :return: a callable that returns with None or raises :class:`TypeError`
    """
    def validate(inst, attribute, value):
        if value not in items:
            raise TypeError("{0} must be one of {1}".format(
                attribute.name, items))
    return validate


@attr.s
class Node(object):
    """
    An object representing a CLB node, which is a unique combination of
    IP-address and port.  Please see section 4.4 (Nodes) of the CLB
    documentation for more information.

    :ivar int id: The ID of the node
    :ivar str address: The IP address of the node
    :ivar int port: The port of the node
    :ivar str type: One of (PRIMARY, SECONDARY).  Defaults to PRIMARY.
    :ivar int weight: Between 1 and 100 inclusive.  Defaults to 1.
    :ivar str condition: One of (ENABLED, DISABLED, DRAINING).  Defaults to
    :ivar str status: "Online"
        ENABLED.
    """
    address = attr.ib(validator=attr.validators.instance_of(string_types))
    port = attr.ib(validator=attr.validators.instance_of(int))
    type = attr.ib(validator=_one_of_validator("PRIMARY", "SECONDARY"),
                   default="PRIMARY")
    weight = attr.ib(validator=attr.validators.instance_of(int), default=1)
    condition = attr.ib(
        validator=_one_of_validator("ENABLED", "DISABLED", "DRAINING"),
        default="ENABLED")
    id = attr.ib(validator=attr.validators.instance_of(int),
                 default=attr.Factory(lambda: randrange(999999)))
    status = attr.ib(validator=attr.validators.instance_of(str),
                     default="ONLINE")
    feed_events = attr.ib(default=[])

    @classmethod
    def from_json(cls, json_blob):
        """
        Create a new node from JSON.

        :param dict json_blob: the JSON dictionary containing node information

        :return: a :class:`Node` object
        :raises: :class:`TypeError` or :class:`ValueError` if the values
            are incorrect.
        """
        # status cannot be in the JSON
        if "status" in json_blob:
            raise ValueError("'status' not allowed in the JSON")

        json_blob['port'] = int(json_blob['port'])
        return Node(**json_blob)

    def as_json(self):
        """
        :return: a JSON dictionary representing the node.
        """
        return dissoc(attr.asdict(self), "feed_events")

    def same_as(self, other):
        """
        :return: `True` if the other node has the same IP address and port
            as this node (but compares nothing else), `False` otherwise.
        """
        return self.address == other.address and self.port == other.port


@attr.s
class CLB(object):
    """
    An object representing a load balancer.  Currently just takes the JSON
    as an attribute, and provides __getitem__ and __setitem__ to access it.

    These should be moved to real attributes as soon as possible.
    """
    _json = attr.ib()
    nodes = attr.ib(default=attr.Factory(list))

    def __getitem__(self, key):
        """
        Convenience function during the full conversion to the object model
        to access JSON keys.
        """
        return self._json[key]

    def __setitem__(self, key, value):
        """
        Convenience function during the full conversion to the object model
        to set JSON keys.
        """
        self._json[key] = value

    def update(self, new_json_dict):
        """
        Convenience function during the full conversion to the object model
        to update all JSON keys.
        """
        self._json.update(new_json_dict)

    def short_json(self):
        """
        :return: a short JSON dict representation of this object to be used
        when listing load balancers.  Does not include the node list, but does
        include a "nodeCount" attribute, even if there are no nodes.
        """
        entries = ('name', 'protocol', 'id', 'port', 'algorithm', 'status',
                   'timeout', 'created', 'virtualIps', 'updated')
        result = dict((entry, self._json[entry]) for entry in entries)
        result['nodeCount'] = len(self.nodes)
        return result

    def full_json(self):
        """
        :return: a longer, detailed JSON dict reprentation of this object that
        includes all the nodes, if there are any present.  Does not include
        a "nodeCount" attribute.
        """
        result = deepcopy(self._json)
        if len(self.nodes) > 0:
            result["nodes"] = [node.as_json() for node in self.nodes]
        return result


@attributes(["keys"])
class BadKeysError(Exception):
    """
    When trying to alter the settings of a load balancer, this exception will
    be raised if you attempt to alter an attribute which doesn't exist.
    """


@attributes(["value", "accepted_values"])
class BadValueError(Exception):
    """
    When trying to alter the settings of a load balancer, this exception will
    be raised if you attempt to set a valid attribute to an invalid setting.
    """


def node_feed_xml(events):
    """
    Return feed of node events
    """
    feed = '<feed xmlns="http://www.w3.org/2005/Atom">{entries}</feed>'
    entry = ('<entry><summary>{summary}</summary>'
             '<updated>{updated}</updated></entry>')
    entries = [entry.format(summary=summary, updated=updated)
               for summary, updated in events]
    return feed.format(entries=''.join(entries))


@attr.s
class RegionalCLBCollection(object):
    """
    A collection of CloudLoadBalancers, in a given region, for a given tenant.
    """
    clock = attr.ib(validator=attr.validators.provides(IReactorTime))
    node_limit = attr.ib(default=25,
                         validator=attr.validators.instance_of(int))
    lbs = attr.ib(default=attr.Factory(dict))
    meta = attr.ib(default=attr.Factory(dict))

    def lb_in_region(self, clb_id):
        """
        Returns true if the CLB ID is registered with our list of load
        balancers.
        """
        return clb_id in self.lbs

    def add_load_balancer(self, lb_info, lb_id):
        """
        Returns response of a newly created load balancer with
        response code 202, and adds the new lb to the store's lbs.
        :param dict lb_info: Configuration for the load balancer.  See
            Openstack docs for creating CLBs.
        :param string lb_id: Unique ID for this load balancer.
        """
        status = "ACTIVE"

        # Loadbalancers metadata is a list object, creating a metadata store
        # so we dont have to deal with the list
        meta = {}
        if "metadata" in lb_info:
            for each in lb_info["metadata"]:
                meta.update({each["key"]: each["value"]})
        self.meta[lb_id] = meta
        log.msg(self.meta)

        if "lb_building" in self.meta[lb_id]:
            status = "BUILD"

        current_timestring = seconds_to_timestamp(self.clock.seconds())
        self.lbs[lb_id] = CLB(load_balancer_example(lb_info, lb_id, status,
                                                    current_timestring),
                              nodes=[Node.from_json(blob)
                                     for blob in lb_info.get("nodes", [])])

        return {'loadBalancer': self.lbs[lb_id].full_json()}, 202

    def _verify_and_update_lb_state(self, lb_id, set_state=True,
                                    current_timestamp=None):
        """
        Based on the current state, the metadata on the lb and the time since the LB has
        been in that state, set the appropriate state in self.lbs
        Note: Reconsider if update metadata is implemented
        """
        current_timestring = seconds_to_timestamp(current_timestamp)
        if self.lbs[lb_id]["status"] == "BUILD":
            self.meta[lb_id]["lb_building"] = self.meta[lb_id]["lb_building"] or 10
            self.lbs[lb_id]["status"] = set_resource_status(
                self.lbs[lb_id]["updated"]["time"],
                self.meta[lb_id]["lb_building"],
                current_timestamp=current_timestamp
            ) or "BUILD"

        elif self.lbs[lb_id]["status"] == "ACTIVE" and set_state:
            if "lb_pending_update" in self.meta[lb_id]:
                self.lbs[lb_id]["status"] = "PENDING-UPDATE"
                log.msg(self.lbs[lb_id]["status"])
            if "lb_pending_delete" in self.meta[lb_id]:
                self.lbs[lb_id]["status"] = "PENDING-DELETE"
            if "lb_error_state" in self.meta[lb_id]:
                self.lbs[lb_id]["status"] = "ERROR"
            self.lbs[lb_id]["updated"]["time"] = current_timestring

        elif self.lbs[lb_id]["status"] == "PENDING-UPDATE":
            if "lb_pending_update" in self.meta[lb_id]:
                self.lbs[lb_id]["status"] = set_resource_status(
                    self.lbs[lb_id]["updated"]["time"],
                    self.meta[lb_id]["lb_pending_update"],
                    current_timestamp=current_timestamp
                ) or "PENDING-UPDATE"

        elif self.lbs[lb_id]["status"] == "PENDING-DELETE":
            self.meta[lb_id]["lb_pending_delete"] = self.meta[lb_id]["lb_pending_delete"] or 10
            self.lbs[lb_id]["status"] = set_resource_status(
                self.lbs[lb_id]["updated"]["time"],
                self.meta[lb_id]["lb_pending_delete"], "DELETED",
                current_timestamp=current_timestamp
            ) or "PENDING-DELETE"
            self.lbs[lb_id]["updated"]["time"] = current_timestring

        elif self.lbs[lb_id]["status"] == "DELETED":
            # see del_load_balancer above for an explanation of this state change.
            self.lbs[lb_id]["status"] = set_resource_status(
                self.lbs[lb_id]["updated"]["time"], 3600, "DELETING-NOW",
                current_timestamp=current_timestamp
            ) or "DELETED"
            if self.lbs[lb_id]["status"] == "DELETING-NOW":
                del self.lbs[lb_id]

    def set_attributes(self, lb_id, kvpairs):
        """
        Sets zero or more attributes on the load balancer object.
        Currently supported attributes include: status.
        """
        supported_keys = ["status"]
        badKeys = []
        for k in kvpairs:
            if k not in supported_keys:
                badKeys.append(k)
        if len(badKeys) > 0:
            raise BadKeysError("Attempt to alter a bad attribute", keys=badKeys)

        if "status" in kvpairs:
            supported_statuses = [
                "ACTIVE", "ERROR", "PENDING_DELETE", "PENDING_UPDATE"
            ]
            s = kvpairs["status"]
            if s not in supported_statuses:
                raise BadValueError(
                    "Unsupported status {0} not one of {1}".format(
                        s, supported_statuses
                    ),
                    value=s, accepted_values=supported_statuses
                )

        self.lbs[lb_id].update(kvpairs)

    def get_load_balancers(self, lb_id):
        """
        Returns the load balancers with the given lb id, with response
        code 200. If no load balancers are found returns 404.
        """
        if lb_id in self.lbs:
            self._verify_and_update_lb_state(lb_id, False, self.clock.seconds())
            log.msg(self.lbs[lb_id]["status"])
            return {'loadBalancer': self.lbs[lb_id].full_json()}, 200
        return not_found_response("loadbalancer"), 404

    def get_node(self, lb_id, node_id):
        """
        Returns the node on the load balancer
        """
        if lb_id in self.lbs:
            self._verify_and_update_lb_state(lb_id, False, self.clock.seconds())

            if self.lbs[lb_id]["status"] == "DELETED":
                return (
                    invalid_resource(
                        "The loadbalancer is marked as deleted.", 410),
                    410)

            for each in self.lbs[lb_id].nodes:
                if node_id == each.id:
                    return {"node": each.as_json()}, 200

            return not_found_response("node"), 404

        return not_found_response("loadbalancer"), 404

    def get_node_feed(self, lb_id, node_id):
        """
        Return load balancer's node's atom feed
        """
        if lb_id not in self.lbs:
            return not_found_xml("Load balancer")

        self._verify_and_update_lb_state(lb_id, False, self.clock.seconds())

        if self.lbs[lb_id]["status"] == "DELETED":
            return lb_deleted_xml()

        for node in self.lbs[lb_id].nodes:
            if node_id == node.id:
                return node_feed_xml(node.feed_events), 200

        return not_found_xml("Node")

    def list_load_balancers(self):
        """
        Returns the list of load balancers with the given tenant id with response
        code 200. If no load balancers are found returns empty list.

        :return: A 2-tuple, containing the HTTP response and code, in that order.
        """
        for lb_id in self.lbs:
            self._verify_and_update_lb_state(lb_id, False, self.clock.seconds())
            log.msg(self.lbs[lb_id]["status"])
        return (
            {'loadBalancers': [lb.short_json() for lb in self.lbs.values()]},
            200)

    def list_nodes(self, lb_id):
        """
        Returns the list of nodes remaining on the load balancer
        """
        if lb_id in self.lbs:
            self._verify_and_update_lb_state(lb_id, False, self.clock.seconds())
            if lb_id not in self.lbs:
                return not_found_response("loadbalancer"), 404

            if self.lbs[lb_id]["status"] == "DELETED":
                return invalid_resource("The loadbalancer is marked as deleted.", 410), 410

            node_list = [node.as_json()
                         for node in self.lbs[lb_id].nodes]

            return {"nodes": node_list}, 200
        else:
            return not_found_response("loadbalancer"), 404

    def _delete_node(self, lb_id, node_id):
        """
        Deletes a node by ID.
        """
        if self.lbs[lb_id].nodes:
            previous_size = len(self.lbs[lb_id].nodes)
            self.lbs[lb_id].nodes[:] = [node
                                        for node in self.lbs[lb_id].nodes
                                        if node.id != node_id]
            return len(self.lbs[lb_id].nodes) < previous_size
        return False

    def delete_node(self, lb_id, node_id):
        """
        Determines whether the node to be deleted exists in the session store,
        deletes the node, and returns the response code.
        """
        current_timestamp = self.clock.seconds()
        if lb_id in self.lbs:

            self._verify_and_update_lb_state(lb_id, False, current_timestamp)

            if self.lbs[lb_id]["status"] != "ACTIVE":
                # Error message verified as of 2015-04-22
                return considered_immutable_error(
                    self.lbs[lb_id]["status"], lb_id)

            self._verify_and_update_lb_state(
                lb_id, current_timestamp=current_timestamp)

            if self._delete_node(lb_id, node_id):
                return None, 202
            else:
                return not_found_response("node"), 404

        return not_found_response("loadbalancer"), 404

    def delete_nodes(self, lb_id, node_ids):
        """
        Bulk-delete multiple LB nodes.
        """
        if not node_ids:
            resp = {
                "message": "Must supply one or more id's to process this request.",
                "code": 400}
            return resp, 400

        if lb_id not in self.lbs:
            return not_found_response("loadbalancer"), 404

        current_timestamp = self.clock.seconds()
        self._verify_and_update_lb_state(lb_id, False, current_timestamp)

        if self.lbs[lb_id]["status"] != "ACTIVE":
            # Error message verified as of 2015-04-22
            resp = {"message": "LoadBalancer is not ACTIVE",
                    "code": 422}
            return resp, 422

        # We need to verify all the deletions up front, and only allow it through
        # if all of them are valid.
        all_ids = [node.id for node in self.lbs[lb_id].nodes]
        non_nodes = set(node_ids).difference(all_ids)
        if non_nodes:
            nodes = ','.join(map(str, non_nodes))
            resp = {
                "validationErrors": {
                    "messages": [
                        "Node ids {0} are not a part of your loadbalancer".format(nodes)
                    ]
                },
                "message": "Validation Failure",
                "code": 400,
                "details": "The object is not valid"}
            return resp, 400

        for node_id in node_ids:
            # It should not be possible for this to fail, since we've already
            # checked that they all exist.
            assert self._delete_node(lb_id, node_id) is True

        self._verify_and_update_lb_state(
            lb_id, current_timestamp=current_timestamp)
        return EMPTY_RESPONSE, 202

    def add_node(self, node_list, lb_id):
        """
        Add one or more nodes to a load balancer.  Fails if one or more of the
        nodes provided has the same address/port as an existing node.  Also
        fails if adding the nodes would exceed the maximum number of nodes on
        the CLB.

        :param list node_list: a `list` of `dict` containing specification for
            nodes

        :return: a `tuple` of (json response as a dict, http status code)
        """
        if lb_id in self.lbs:
            current_timestamp = self.clock.seconds()
            self._verify_and_update_lb_state(lb_id, False, current_timestamp)

            if self.lbs[lb_id]["status"] != "ACTIVE":
                return considered_immutable_error(
                    self.lbs[lb_id]["status"], lb_id)

            nodes = [Node.from_json(blob) for blob in node_list]

            for existing_node in self.lbs[lb_id].nodes:
                for new_node in nodes:
                    if existing_node.same_as(new_node):
                        resource = invalid_resource(
                            "Duplicate nodes detected. One or more nodes "
                            "already configured on load balancer.", 413)
                        return (resource, 413)

            # If there were no duplicates
            new_nodeCount = len(self.lbs[lb_id].nodes) + len(nodes)
            if new_nodeCount <= self.node_limit:
                self.lbs[lb_id].nodes = self.lbs[lb_id].nodes + nodes
            else:
                resource = invalid_resource(
                    "Nodes must not exceed {0} "
                    "per load balancer.".format(self.node_limit), 413)
                return (resource, 413)

            self._verify_and_update_lb_state(
                lb_id, current_timestamp=current_timestamp)
            return {"nodes": [node.as_json() for node in nodes]}, 202

        return not_found_response("loadbalancer"), 404

    def update_node(self, lb_id, node_id, node_updates):
        """
        Update the weight, condition, or type of a single node.  The IP, port,
        status, and ID are immutable, and attempting to change them will cause
        a 400 response to be returned.

        All success and error behavior verified as of 2016-06-16.

        :param str lb_id: the load balancer ID
        :param str node_id: the node ID to update
        :param dict node_updates: The JSON dictionary containing node
            attributes to update
        :param current_timestamp: What the current time is

        :return: a `tuple` of (json response as a dict, http status code)
        """
        feed_summary = (
            "Node successfully updated with address: '{address}', port: '{port}', "
            "weight: '{weight}', condition: '{condition}'")
        # first, store whether address and port were provided - if they were
        # that's a validation error not a schema error
        things_wrong = dict([(k, True) for k in ("address", "port", "id")
                             if k in node_updates])
        node_updates = dict([(k, v) for k, v in node_updates.items()
                             if k not in ("address", "port")])
        # use the Node.from_json to check the schema
        try:
            Node.from_json(dict(address="1.1.1.1", port=80, **node_updates))
        except (TypeError, ValueError):
            return invalid_json_schema()

        # handle the possible validation (as opposed to schema) errors
        if not 1 <= node_updates.get('weight', 1) <= 100:
            things_wrong["weight"] = True
        if things_wrong:
            return updating_node_validation_error(**things_wrong)

        # Now, finally, check if the LB exists and node exists
        if lb_id in self.lbs:
            self._verify_and_update_lb_state(lb_id, False, self.clock.seconds())

            if self.lbs[lb_id]["status"] != "ACTIVE":
                return considered_immutable_error(
                    self.lbs[lb_id]["status"], lb_id)

            for i, node in enumerate(self.lbs[lb_id].nodes):
                if node.id == node_id:
                    params = attr.asdict(node)
                    params.update(node_updates)
                    self.lbs[lb_id].nodes[i] = Node(**params)
                    self.lbs[lb_id].nodes[i].feed_events.append(
                        (feed_summary.format(**params),
                         seconds_to_timestamp(self.clock.seconds())))
                    return ("", 202)

            return node_not_found()

        return loadbalancer_not_found()

    def del_load_balancer(self, lb_id):
        """
        Returns response for a load balancer
         is in building status for 20
        seconds and response code 202, and adds the new lb to ``self.lbs``.
        A loadbalancer, on delete, goes into PENDING-DELETE and remains in DELETED
        status until a nightly job(maybe?)
        """
        if lb_id in self.lbs:
            current_timestamp = self.clock.seconds()

            if self.lbs[lb_id]["status"] == "PENDING-DELETE":
                msg = ("Must provide valid load balancers: {0} are immutable and "
                       "could not be processed.".format(lb_id))
                # Dont doubt this to be 422, it is 400!
                return invalid_resource(msg, 400), 400

            self._verify_and_update_lb_state(lb_id, True, current_timestamp)

            if any([self.lbs[lb_id]["status"] == "ACTIVE",
                    self.lbs[lb_id]["status"] == "ERROR",
                    self.lbs[lb_id]["status"] == "PENDING-UPDATE"]):
                del self.lbs[lb_id]
                return EMPTY_RESPONSE, 202

            if self.lbs[lb_id]["status"] == "PENDING-DELETE":
                return EMPTY_RESPONSE, 202

            if self.lbs[lb_id]["status"] == "DELETED":
                self._verify_and_update_lb_state(
                    lb_id, current_timestamp=current_timestamp)
                msg = "Must provide valid load balancers: {0} could not be found.".format(lb_id)
                # Dont doubt this to be 422, it is 400!
                return invalid_resource(msg, 400), 400

        return not_found_response("loadbalancer"), 404


@attributes(["clock",
             Attribute("regional_collections", default_factory=dict)])
class GlobalCLBCollections(object):
    """
    A :obj:`GlobalCLBCollections` is a set of all the
    :obj:`RegionalCLBCollection` objects owned by a given tenant.  In other
    words, all the objects that a single tenant owns globally in a
    cloud load balancer service.
    """

    def collection_for_region(self, region_name):
        """
        Get a :obj:`RegionalCLBCollection` for the region identified by the
        given name.
        """
        if region_name not in self.regional_collections:
            self.regional_collections[region_name] = (
                RegionalCLBCollection(self.clock)
            )
        return self.regional_collections[region_name]



"""
Twisted Application plugin for Mimic
"""
from twisted.application.strports import service
from twisted.application.service import MultiService
from twisted.python import usage
from mimic.core import MimicCore
from mimic.resource import MimicRoot, get_site
from twisted.internet.task import Clock


class Options(usage.Options):
    """
    Options for Mimic
    """
    optParameters = [['listen', 'l', '8900', 'The endpoint to listen on.']]
    optFlags = [['realtime', 'r',
                 'Make mimic advance time as real time advances; '
                 'disable the "tick" endpoint.'],
                ['verbose', 'v',
                 'Log more verbosely: include full requests and responses.']]


def makeService(config):
    """
    Set up the otter-api service.
    """
    s = MultiService()
    if config['realtime']:
        from twisted.internet import reactor as clock
    else:
        clock = Clock()
    core = MimicCore.fromPlugins(clock)
    root = MimicRoot(core, clock)
    site = get_site(root.app.resource(), logging=bool(config['verbose']))
    service(config['listen'], site).setServiceParent(s)
    return s



"""
Mailgun object storage
"""
import time
from characteristic import attributes, Attribute


@attributes(["message_id", "to", "msg_from", "subject", "body",
             Attribute("custom_headers", default_factory=dict)])
class Message(object):
    """
    A :obj:`Message` is a representation of an email in Mailgun.
    It can produce JSON-serializable objects for various pieces of
    state that are required for API responses.
    """

    static_defaults = {
        "tags": [],
        "delivery-status": {
            "message": "",
            "code": 0,
            "description": None,
            "session-seconds": 1.114408016204834
        },
        "envelope": {
            "transport": "smtp",
            "sending-ip": "127.0.0.1",
        },
        "recipient-domain": "mailgun.com",
        "id": "mimic-LCZuENBlS0iWjs-yBpNJaQ",
        "campaigns": [],
        "user-variables": {},
        "flags": {
            "is-routed": None,
            "is-authenticated": True,
            "is-system-test": False,
            "is-test-mode": False
        },
        "log-level": "info",
        "timestamp": time.time(),
        "message": {
            "headers": {},
            "attachments": [],
            "recipients": [],
            "size": 0
        },
        "recipient": None,
        "event": "delivered"
    }

    def generate_events(self):
        """
        Long-form JSON-serializable object representation of this message, as
        returned by a GET on this individual message.
        """
        template = self.static_defaults.copy()
        template.update({
            "envelope": {
                "sender": self.msg_from,
                "targets": self.to
            },
            "message": {
                "headers": {
                    "to": self.to,
                    "message-id": self.message_id,
                    "from": self.msg_from,
                    "subject": self.subject
                },
                "recipients": [self.to],
                "recipient": self.to
            }
        })
        return template


@attributes([Attribute("message_store", default_factory=list)])
class MessageStore(object):
    """
    A collection of message objects.
    """

    def add_to_message_store(self, **attributes):
        """
        Create a new Message object and add it to the
        :obj: `message_store`
        """
        msg = Message(**attributes)
        self.message_store.append(msg)
        return

    def list_messages(self, filter_by=None):
        """
        List all the messages.

        :param str filter_by: supports filtering the List by
        `to` addresses only currently.
        """
        to_be_listed = self.message_store
        if filter_by:
            events = [msg.generate_events()
                      for msg in to_be_listed if msg.to in filter_by]
        else:
            events = [msg.generate_events() for msg in to_be_listed]
        return {
            "items": events,
            "paging": {
                "next": "http://i-am-a-fake-link-to-nothing",
                "last": "http://i-am-a-fake-link-to-nothing",
                "first": "http://i-am-a-fake-link-to-nothing=",
                "previous": "http://i-am-a-fake-link-to-nothing=="
            }}

    def filter_message_by_to_address(self, to_address):
        """
        Retrieve a :obj:`Message` object by its `to` address.
        """
        for each_msg in self.message_store:
            if each_msg.to == to_address[0]:
                return each_msg



# -*- test-case-name: mimic.test.test_loadbalancer -*-
"""
Defines the control plane API endpoints for the Cloudfeeds Plugin.
"""
from uuid import uuid4
from six import text_type
from zope.interface import implementer
from twisted.plugin import IPlugin
from mimic.catalog import Endpoint, Entry
from mimic.imimic import IAPIMock
from mimic.rest.mimicapp import MimicApp


from characteristic import attributes


@implementer(IAPIMock, IPlugin)
class CloudFeedsApi(object):
    """
    This class registers the cloud feeds API in the service catalog.
    """
    def __init__(self, regions=["ORD"]):
        """
        Configures the API with a list of regions.

        :param list regions: If provided, a list of strings, each providing a
            name for a region (e.g., "ORD").  If not specified, defaults to
            ["ORD"].
        """
        self._regions = regions

    def catalog_entries(self, tenant_id):
        """
        Returns a list of cloud feeds entries.  Note that these are not
        cloud feeds product endpoints; this is one step removed from those.
        You'll need to GET from one of these URLs to see the catalog of
        product endpoints.
        """
        return [
            Entry(tenant_id, "rax:feeds", "cloudFeeds",
                  [
                      Endpoint(tenant_id, region, text_type(uuid4()))
                      for region in self._regions
                  ])
        ]

    def resource_for_region(self, region, uri_prefix, session_store):
        """
        Get an :obj:`twisted.web.iweb.IResource` for the given URI prefix;
        implement :obj:`IAPIMock`.
        """
        lb_region = CloudFeedsRegion(self, uri_prefix, session_store, region)
        return lb_region.app.resource()


@implementer(IAPIMock, IPlugin)
@attributes(["cf_api"])
class CloudFeedsControlApi(object):
    """
    This class registers the load balancer controller API in the service
    catalog.
    """
    def catalog_entries(self, tenant_id):
        """
        Cloud feeds controller endpoints.
        """
        return [
            Entry(
                tenant_id, "rax:feeds", "cloudFeedsControl",
                [
                    Endpoint(tenant_id, region, text_type(uuid4()))
                    for region in self.cf_api._regions
                ]
            )
        ]

    def resource_for_region(self, region, uri_prefix, session_store):
        """
        Get an :obj:`twisted.web.iweb.IResource` for the given URI prefix;
        implement :obj:`IAPIMock`.
        """
        cfc_region = CloudFeedsControlRegion(
            api_mock=self, uri_prefix=uri_prefix,
            session_store=session_store, region=region
        )
        return cfc_region.app.resource()


@attributes(["api_mock", "uri_prefix", "session_store", "region"])
class CloudFeedsControlRegion(object):
    """
    Klein routes for cloud feed's control API within a particular region.
    """

    app = MimicApp()


class CloudFeedsRegion(object):
    """
    Klein routes for cloud feeds API methods within a particular region.
    """

    app = MimicApp()

    def __init__(self, api_mock, uri_prefix, session_store, region_name):
        """
        Fetches the cloud feeds id for a failure, invalid scenarios, etc.
        """
        self.uri_prefix = uri_prefix
        self.region_name = region_name
        self._api_mock = api_mock
        self._session_store = session_store



"""
Tests for :mod:`nova_api` and :mod:`nova_objects`.
"""

from twisted.trial.unittest import SynchronousTestCase

from mimic.test.helpers import json_request, request
from mimic.rest.nova_api import NovaApi, NovaControlApi
from mimic.test.fixtures import APIMockHelper


class KeyPairTests(SynchronousTestCase):
    """
    Tests for keypairs
    """

    def setUp(self):
        """
        Create a :obj:`MimicCore` with :obj:`NovaApi` as the only plugin
        """
        nova_api = NovaApi(["ORD", "MIMIC"])
        self.helper = APIMockHelper(
            self, [nova_api, NovaControlApi(nova_api=nova_api)]
        )
        self.root = self.helper.root
        self.clock = self.helper.clock
        self.uri = self.helper.uri
        self.create_keypair_response, self.create_keypair_response_body = (
            self.create_keypair())
        self.keypair_name = self.create_keypair_response_body[
            'keypair']['name']

    def create_keypair(self, kp_body=None):
        if kp_body is None:
            kp_body = {
                "keypair": {
                    "name": "setUP_test_lp",
                    "public_key": "ssh-rsa testkey/"
                }
            }

        resp, body = self.successResultOf(json_request(
            self, self.helper.root, "POST", self.helper.uri + '/os-keypairs',
            kp_body
        ))
        return resp, body

    def get_keypairs_list(self):
        resp, body = self.successResultOf(json_request(
            self, self.helper.root, "GET", self.helper.uri + '/os-keypairs'
        ))
        return resp, body

    def test_create_keypair(self):
        kp_test_body = {
            "keypair": {
                "name": "test_lp",
                "public_key": "ssh-rsa testkey/"
            }
        }
        resp, body = self.create_keypair(kp_test_body)

        self.assertEqual(resp.code, 200)
        self.assertEqual(body['keypair']['name'],
                         kp_test_body['keypair']['name'])
        self.assertTrue(len(body['keypair']['fingerprint']) > 1)
        self.assertTrue(len(body['keypair']['user_id']) > 1)

    def test_error_create_keypair(self):
        test_error_body = "{{a]"
        resp, body = self.create_keypair(test_error_body)
        self.assertEqual(resp.code, 400)
        self.assertSubstring("Malformed", str(body))

    def test_list_keypair(self):
        resp, body = self.get_keypairs_list()
        self.assertEqual(resp.code, 200)
        self.assertEqual(body['keypairs'][0]['keypair']
                         ['name'], self.keypair_name)

    def test_delete_keypair(self):
        resp = self.successResultOf(request(
            self, self.helper.root, "DELETE", self.helper.uri +
            '/os-keypairs/' + self.keypair_name
        ))

        self.assertEqual(resp.code, 202)
        resp, body = self.get_keypairs_list()
        self.assertNotIn(self.keypair_name, str(body['keypairs']))

    def test_error_delete_keypair(self):
        resp = self.successResultOf(request(
            self, self.helper.root, "DELETE", self.helper.uri +
            '/os-keypairs/keydoesntexist'
        ))

        self.assertEqual(resp.code, 404)



"""
Unit tests for the Rackspace RackConnect V3 API.
"""
import json
from random import randint
from uuid import uuid4

from six import text_type

from twisted.trial.unittest import SynchronousTestCase
from mimic.test.fixtures import APIMockHelper
from mimic.rest.rackconnect_v3_api import (
    LoadBalancerPool, LoadBalancerPoolNode, RackConnectV3)
from mimic.test.helpers import json_request, request_with_content


class _IsString(object):

    """
    Helper class to be used when checking equality when you don't what the ID
    is but you want to check that it's an ID
    """

    def __eq__(self, other):
        """
        Returns true if the other is a string
        """
        return isinstance(other, basestring)


class LoadBalancerObjectTests(SynchronousTestCase):

    """
    Tests for :class:`LoadBalancerPool` and :class:`LoadBalancerPoolNode`
    """

    def setUp(self):
        self.pool = LoadBalancerPool(id=text_type("pool_id"),
                                     virtual_ip=text_type("10.0.0.1"))
        for i in range(10):
            self.pool.nodes.append(
                LoadBalancerPoolNode(id=text_type("node_{0}".format(i)),
                                     created="2000-01-01T00:00:00Z",
                                     load_balancer_pool=self.pool,
                                     updated=None,
                                     cloud_server=text_type(
                                         "server_{0}".format(i))))

    def test_LBPoolNode_short_json(self):
        """
        Valid JSON response (as would be displayed when listing nodes) is
        produced by :func:`LoadBalancerPoolNode.short_json`
        """
        self.assertEqual(
            {
                "id": "node_0",
                "created": "2000-01-01T00:00:00Z",
                "updated": None,
                "load_balancer_pool": {
                    "id": "pool_id"
                },
                "cloud_server": {
                    "id": "server_0"
                },
                "status": "ACTIVE",
                "status_detail": None
            },
            self.pool.nodes[0].short_json())

    def test_LBPoolNode_update(self):
        """
        Updating the status changes the 'now', 'status', and 'status_detail'
        attributes.
        """
        self.pool.nodes[0].update(now="2000-01-02T00:00:00Z",
                                  status="DISABLED",
                                  status_detail="Broken.")
        self.assertEqual(
            {
                "id": "node_0",
                "created": "2000-01-01T00:00:00Z",
                "updated": "2000-01-02T00:00:00Z",
                "load_balancer_pool": {
                    "id": "pool_id"
                },
                "cloud_server": {
                    "id": "server_0"
                },
                "status": "DISABLED",
                "status_detail": "Broken."
            },
            self.pool.nodes[0].short_json())

    def test_LBPool_short_json(self):
        """
        Valid JSON response (as would be displayed when listing pools or
        getting pool details) is produced by :func:`LoadBalancerPool.as_json`.
        """
        self.assertEqual(
            {
                "id": "pool_id",
                "name": "default",
                "node_counts": {
                    "cloud_servers": 10,
                    "external": 0,
                    "total": 10
                },
                "port": 80,
                "virtual_ip": "10.0.0.1",
                "status": "ACTIVE",
                "status_detail": None
            },
            self.pool.as_json())

    def test_LBPool_find_nodes_by_id(self):
        """
        A node can be retrieved by its ID.
        """
        self.assertIs(self.pool.nodes[5], self.pool.node_by_id("node_5"))

    def test_LBPool_find_nodes_by_server_id(self):
        """
        A node can be retrieved by its cloud server ID.
        """
        self.assertIs(self.pool.nodes[3],
                      self.pool.node_by_cloud_server("server_3"))


class RackConnectTestMixin(object):

    """
    Mixin object that provides some nice utilities
    """

    def setUp(self):
        """
        Create a :obj:`MimicCore` with :obj:`RackConnectV3` as the only plugin
        """
        super(RackConnectTestMixin, self).setUp()
        self.rcv3 = RackConnectV3()
        self.helper = APIMockHelper(self, [self.rcv3])
        self.pool_id = self.get_lb_ids()[0][0]

    def get_lb_ids(self):
        """
        Helper function to get the load balancer ids per region
        """
        _, resp_jsons = zip(*[
            self.successResultOf(json_request(
                self, self.helper.root, "GET",
                self.helper.get_service_endpoint("rackconnect", region)
                + "/load_balancer_pools"))
            for region in self.rcv3.regions])

        lb_ids = [[lb['id'] for lb in lbs] for lbs in resp_jsons]
        return lb_ids

    def request_with_content(self, method, relative_uri, **kwargs):
        """
        Helper function that makes a request and gets the non-json content.
        """
        return request_with_content(self, self.helper.root, method,
                                    self.helper.uri + relative_uri, **kwargs)

    def json_request(self, method, relative_uri, **kwargs):
        """
        Helper function that makes a request and gets the json content.
        """
        return json_request(self, self.helper.root, method,
                            self.helper.uri + relative_uri, **kwargs)


class LoadbalancerPoolAPITests(RackConnectTestMixin, SynchronousTestCase):

    """
    Tests for the LoadBalancerPool API
    """

    def test_list_pools_default_one(self):
        """
        Verify the JSON response from listing all load balancer pools.
        By default, all tenants have one load balancer pool.
        """
        response, response_json = self.successResultOf(
            self.json_request("GET", "/load_balancer_pools"))
        self.assertEqual(200, response.code)
        self.assertEqual(['application/json'],
                         response.headers.getRawHeaders('content-type'))
        self.assertEqual(1, len(response_json))

        pool_json = response_json[0]
        # has the right JSON
        self.assertTrue(all(
            attr.name in pool_json
            for attr in LoadBalancerPool.characteristic_attributes
            if attr.name != "nodes"))
        # Generated values
        self.assertTrue(all(
            pool_json.get(attr.name)
            for attr in LoadBalancerPool.characteristic_attributes
            if attr.name not in ("nodes", "status_detail")))

        self.assertEqual(
            {
                "cloud_servers": 0,
                "external": 0,
                "total": 0
            },
            pool_json['node_counts'],
            "Pool should start off with no members.")

    def test_different_regions_same_tenant_different_pools(self):
        """
        The same tenant has different pools in different regions, default of 1
        pool in each.
        """
        self.rcv3 = RackConnectV3(regions=["ORD", "DFW"])
        self.helper = APIMockHelper(self, [self.rcv3])
        lb_ids = self.get_lb_ids()
        self.assertEqual(1, len(lb_ids[0]))
        self.assertEqual(1, len(lb_ids[1]))
        self.assertNotEqual(set(lb_ids[0]), set(lb_ids[1]))

    def test_default_multiple_pools(self):
        """
        If ``default_pools`` is passed to :class:`RackConnectV3`, multiple
        load balancer pools will be created per tenant per region
        """
        self.rcv3 = RackConnectV3(regions=["ORD", "DFW"], default_pools=2)
        self.helper = APIMockHelper(self, [self.rcv3])
        lb_ids = self.get_lb_ids()
        self.assertEqual(2, len(lb_ids[0]))
        self.assertEqual(2, len(lb_ids[1]))
        self.assertNotEqual(set(lb_ids[0]), set(lb_ids[1]))

    def test_get_pool_on_success(self):
        """
        Validate the JSON response of getting a single pool on an existing
        pool.
        """
        _, pool_list_json = self.successResultOf(
            self.json_request("GET", "/load_balancer_pools"))
        pool = pool_list_json[0]

        pool_details_response, pool_details_json = self.successResultOf(
            self.json_request("GET",
                              "/load_balancer_pools/{0}".format(pool['id'])))

        self.assertEqual(200, pool_details_response.code)
        self.assertEqual(
            ['application/json'],
            pool_details_response.headers.getRawHeaders('content-type'))
        self.assertEqual(pool, pool_details_json)

    def test_get_pool_400_non_uuid_pool_id(self):
        """
        Getting pool on a non-uuid pool id returns a 400.
        """
        response, content = self.successResultOf(
            self.request_with_content("GET", "/load_balancer_pools/123"))

        self.assertEqual(400, response.code)

    def test_get_pool_404_non_existant_pool(self):
        """
        Getting pool on a non-existant pool returns a 404.
        """
        random_pool_id = text_type(uuid4())
        response, content = self.successResultOf(
            self.request_with_content("GET",
                                      "/load_balancer_pools/{0}".format(random_pool_id)))

        self.assertEqual(404, response.code)
        self.assertEqual("Load Balancer Pool {0} does not exist".format(random_pool_id), content)

    def _get_add_nodes_json(self):
        """
        Helper function to generate bulk add nodes JSON given the lbs on
        the tenant and region
        """
        return [
            {"cloud_server": {"id": "{0}".format(randint(0, 9))},
             "load_balancer_pool": {"id": pool_id}}
            for pool_id in self.get_lb_ids()[0]
        ]

    def _get_custom_add_nodes_json(self, node_and_pool_ids):
        """
        Helper function to generate bulk add nodes JSON given values for
        node and pool ids.
        node_and_pool_ids: list of tuples of corresponding node and
        pool ids.
        """
        return [
            {"cloud_server": {"id": "{0}".format(each[0])},
             "load_balancer_pool": {"id": each[1]}}
            for each in node_and_pool_ids
        ]

    def _check_added_nodes_result(self, seconds, add_json, results_json):
        """
        Helper function to add some servers to the pools, and check that the
        results reflect the added nodes
        """
        self.assertEqual(len(add_json), len(results_json))

        # sort function by server ID then load balancer pool ID
        def cmp_key_function(dictionary):
            "{0}_{1}".format(dictionary['cloud_server']['id'],
                             dictionary['load_balancer_pool']['id'])

        add_json = sorted(add_json, key=cmp_key_function)
        results_json = sorted(results_json, key=cmp_key_function)

        # Can't construct the whole thing, because the IDs are random, so
        # compare some parts
        for i, add_blob in enumerate(add_json):
            result = results_json[i]
            expected = {
                "id": _IsString(),
                "cloud_server": add_blob['cloud_server'],
                "load_balancer_pool": add_blob['load_balancer_pool'],
                "created": "1970-01-01T00:00:{0:02}Z".format(seconds),
                "status": "ACTIVE",
                "status_detail": None,
                "updated": None
            }
            self.assertEqual(expected, result)

    def test_add_bulk_pool_nodes_success_response(self):
        """
        Adding multiple pool nodes successfully results in a 200 with the
        correct node detail responses
        """
        self.helper.clock.advance(50)
        add_data = self._get_add_nodes_json()
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(201, response.code)
        self._check_added_nodes_result(50, add_data, resp_json)

    def test_add_bulk_pool_nodes_when_pool_id_is_non_uuid(self):
        """
        Adding multiple pool nodes results in a 400 if one of the pool_ids
        is not a uuid4
        """
        self.helper.clock.advance(50)
        node_and_pool_ids = [(text_type(uuid4()), 122),
                             (text_type(uuid4()), text_type(uuid4()))]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(400, response.code)

    def test_add_bulk_pool_nodes_to_single_non_existant_pool_id_1(self):
        """
        Adding a pool node to a non-existant pool_id using add bulk node
        api call results in a 409 with respective error message
        """
        self.helper.clock.advance(50)
        pool_id = text_type(uuid4())
        node_and_pool_ids = [(text_type(uuid4()), pool_id)]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(409, response.code)
        self.assertEqual("Load Balancer Pool {0} does not exist".format(pool_id),
                         resp_json["errors"][0])

    def test_add_bulk_pool_nodes_to_single_non_existant_pool_id_2(self):
        """
        Adding multiple pool nodes to a non-existant pool_id results in a 409
        with respective error message
        """
        self.helper.clock.advance(50)
        pool_id = text_type(uuid4())
        node_and_pool_ids = [(text_type(uuid4()), pool_id),
                             (text_type(uuid4()), pool_id)]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(409, response.code)
        self.assertEqual("Load Balancer Pool {0} does not exist".format(pool_id),
                         resp_json["errors"][0])
        self.assertEqual("Load Balancer Pool {0} does not exist".format(pool_id),
                         resp_json["errors"][1])

    def test_add_bulk_pool_nodes_to_multiple_non_existant_pool_ids(self):
        """
        Adding multiple pool nodes to multiple non-existant pool_id results in a 409
        with respective error messages
        """
        self.helper.clock.advance(50)
        node_and_pool_ids = [(text_type(uuid4()), text_type(uuid4())),
                             (text_type(uuid4()), text_type(uuid4()))]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(409, response.code)
        self.assertEqual(len(resp_json['errors']), 2)

    def test_add_bulk_pool_nodes_for_an_existing_node(self):
        """
        Adding an existing node to a pool_id results in a 409
        with respective error messages
        """
        self.helper.clock.advance(50)
        server_id = text_type(uuid4())

        # add node to lb_pool
        node_and_pool_ids = [(server_id, self.get_lb_ids()[0][0])]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(201, response.code)

        # re-adding the node
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(409, response.code)
        self.assertEqual("Cloud Server {0} is already a member of "
                         "Load Balancer Pool {1}".format(server_id, self.get_lb_ids()[0][0]),
                         resp_json["errors"][0])

    def test_add_bulk_pool_nodes_multiple_errors(self):
        """
        Adding an existing node to a pool_id. Then adding the node to an exiting
        and non-existing pool id results in a 409 with respective error messages
        """
        self.helper.clock.advance(50)

        # add node to a valid lb pool
        server_id = text_type(uuid4())
        node_and_pool_ids = [(server_id, self.get_lb_ids()[0][0])]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(201, response.code)

        # get the node count for the lb pool
        _, list_json = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))

        # re-add the node to a valid lb pool id as well to a non-existant pool_id
        node_and_pool_ids.append((server_id, text_type(uuid4())))
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(409, response.code)
        self.assertEqual(len(resp_json["errors"]), 2)

        # ensure no new nodes were added to the pool
        _, list_json_again = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))
        self.assertEqual(len(list_json_again), len(list_json))

    def test_add_bulk_pool_nodes_errors_with_no_node_added(self):
        """
        Add bulk pool nodes fails with a 409 if even one of the cloud servers
        or pool_ids in the request causes an error.
        """
        self.helper.clock.advance(50)
        # get the node count for the lb pool
        _, list_json = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))
        server_id = text_type(uuid4())
        random_pool_id = text_type(uuid4())
        node_and_pool_ids = [(server_id, self.get_lb_ids()[0][0]),
                             (server_id, random_pool_id)]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(409, response.code)
        self.assertEqual("Load Balancer Pool {0} does not exist".format(random_pool_id),
                         resp_json["errors"][0])
        # ensure no new nodes were added to the pool
        _, list_json_again = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))
        self.assertEqual(len(list_json_again), len(list_json))

    def test_add_bulk_pool_nodes_then_list(self):
        """
        Adding multiple pool nodes successfully means that the next time nodes
        are listed those nodes are listed.
        """
        self.helper.clock.advance(50)
        add_data = self._get_add_nodes_json()
        add_response, _ = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(201, add_response.code)

        _, list_json = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))
        self._check_added_nodes_result(50, add_data, list_json)

    def test_remove_bulk_pool_nodes_success(self):
        """
        Removing multiple pool nodes successfully results in a 204 with the
        correct node detail responses
        """
        server_data = self._get_add_nodes_json()

        # add first
        resp, body = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=server_data))

        # ensure the node has been added
        _, list_json = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))
        self.assertEqual(1, len(list_json))

        # delete
        response, _ = self.successResultOf(self.request_with_content(
            "DELETE", "/load_balancer_pools/nodes",
            body=json.dumps(server_data)))
        self.assertEqual(204, response.code)

        # ensure there are 0
        _, list_json = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))
        self.assertEqual(0, len(list_json))

    def test_remove_bulk_pool_nodes_when_pool_id_is_non_uuid(self):
        """
        Removing multiple pool nodes results in a 400 if one of the pool_ids
        is not a uuid4
        """
        node_and_pool_ids = [(text_type(uuid4()), 122)]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "DELETE", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(400, response.code)

    def test_remove_bulk_pool_nodes_to_single_non_existant_pool_id_1(self):
        """
        Removing a pool node from a non-existant pool_id results in a 409
        with respective error message
        """
        server_id = text_type(uuid4())
        node_and_pool_ids = [(server_id, self.get_lb_ids()[0][0])]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(201, response.code)
        pool_id = text_type(uuid4())
        node_and_pool_ids = [(server_id, pool_id)]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "DELETE", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(409, response.code)
        self.assertEqual("Load Balancer Pool {0} does not exist".format(pool_id),
                         resp_json["errors"][0])

    def test_remove_bulk_pool_nodes_to_single_non_existant_pool_id_2(self):
        """
        Removing multiple pool nodes from a non-existant pool_id results in a 409
        with respective error message
        """
        pool_id = text_type(uuid4())
        node_and_pool_ids = [(text_type(uuid4()), pool_id),
                             (text_type(uuid4()), pool_id)]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "DELETE", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(409, response.code)
        self.assertEqual("Load Balancer Pool {0} does not exist".format(pool_id),
                         resp_json["errors"][0])

    def test_remove_bulk_pool_nodes_to_multiple_non_existant_pool_ids(self):
        """
        Removing multiple pool nodes from multiple non-existant pool_id results in a 409
        with respective error messages
        """
        node_and_pool_ids = [(text_type(uuid4()), text_type(uuid4())),
                             (text_type(uuid4()), text_type(uuid4()))]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(409, response.code)
        self.assertEqual(len(resp_json['errors']), 2)

    def test_remove_bulk_pool_nodes_given_a_non_existing_node(self):
        """
        Removing an already deleted node from a pool_id results in a 409
        with respective error messages
        """
        server_id = text_type(uuid4())
        node_and_pool_ids = [(server_id, self.get_lb_ids()[0][0])]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(201, response.code)
        response, _ = self.successResultOf(self.request_with_content(
            "DELETE", "/load_balancer_pools/nodes",
            body=json.dumps(add_data)))
        self.assertEqual(204, response.code)

        # get the node count for the lb pool
        _, list_json = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))

        response, resp_json = self.successResultOf(self.request_with_content(
            "DELETE", "/load_balancer_pools/nodes",
            body=json.dumps(add_data)))
        self.assertEqual(409, response.code)
        resp = json.loads(resp_json)
        self.assertEqual("Cloud Server {0} is not a member of "
                         "Load Balancer Pool {1}".format(server_id, self.get_lb_ids()[0][0]),
                         resp["errors"][0])

        # ensure no other nodes were deleted from the pool
        _, list_json_again = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))
        self.assertEqual(len(list_json_again), len(list_json))

    def test_remove_bulk_pool_nodes_multiple_errors(self):
        """
        Removing bulk pool nodes fails with a 409 if even one of the cloud servers
        or pool_ids in the request causes an error.
        """
        server_id = text_type(uuid4())
        random_pool_id = text_type(uuid4())

        # get the node count for the lb pool
        _, list_json = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))

        node_and_pool_ids = [(server_id, self.get_lb_ids()[0][0]),
                             (server_id, random_pool_id)]
        add_data = self._get_custom_add_nodes_json(node_and_pool_ids)
        response, resp_json = self.successResultOf(self.json_request(
            "POST", "/load_balancer_pools/nodes", body=add_data))
        self.assertEqual(409, response.code)
        self.assertEqual("Load Balancer Pool {0} does not exist".format(random_pool_id),
                         resp_json["errors"][0])

        # ensure no nodes were deleted from the pool
        _, list_json_again = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))
        self.assertEqual(len(list_json_again), len(list_json))


class LoadbalancerPoolNodesAPITests(RackConnectTestMixin,
                                    SynchronousTestCase):

    """
    Tests for the LoadBalancerPool API for getting and updating nodes
    """

    def test_get_pool_404_invalid_pool_nodes(self):
        """
        Getting nodes on a non-existant pool returns a 404.
        """
        random_pool_id = text_type(uuid4())
        response, content = self.successResultOf(self.request_with_content(
            "GET", "/load_balancer_pools/{0}/nodes".format(random_pool_id)))

        self.assertEqual(404, response.code)
        self.assertEqual("Load Balancer Pool {0} does not exist".format(random_pool_id),
                         content)

    def test_get_pool_nodes_empty(self):
        """
        Getting nodes for an empty existing load balancer returns a 200 with
        no nodes
        """
        response, json_content = self.successResultOf(self.json_request(
            "GET", "/load_balancer_pools/{0}/nodes".format(self.pool_id)))
        self.assertEqual(200, response.code)
        self.assertEqual(json_content, [])

    def test_get_pool_nodes_details_unimplemented(self):
        """
        Getting pool nodes details is currently unimplemented
        """
        response, content = self.successResultOf(self.request_with_content(
            "GET",
            "/load_balancer_pools/{0}/nodes/details".format(self.pool_id)))
        self.assertEqual(501, response.code)

    def test_add_pool_node_unimplemented(self):
        """
        Adding a single pool node is currently unimplemented
        """
        response, content = self.successResultOf(self.request_with_content(
            "POST", "/load_balancer_pools/{0}/nodes".format(self.pool_id),
            body=json.dumps({
                "cloud_server": {"id": "d95ae0c4-6ab8-4873-b82f-f8433840cff2"}
            })))
        self.assertEqual(501, response.code)

    def test_get_pool_node_unimplemented(self):
        """
        Getting information a single pool node is currently unimplemented
        """
        response, content = self.successResultOf(self.request_with_content(
            "GET", "/load_balancer_pools/{0}/nodes/1".format(self.pool_id)
        ))
        self.assertEqual(501, response.code)

    def test_remove_pool_node_unimplemented(self):
        """
        Removing a single pool node is currently unimplemented
        """
        response, content = self.successResultOf(self.request_with_content(
            "DELETE", "/load_balancer_pools/{0}/nodes/1".format(self.pool_id)
        ))
        self.assertEqual(501, response.code)

    def test_get_pool_node_details_unimplemented(self):
        """
        Getting detailed information on a single pool node is currently
        unimplemented
        """
        response, content = self.successResultOf(self.request_with_content(
            "GET",
            "/load_balancer_pools/{0}/nodes/1/details".format(self.pool_id)
        ))
        self.assertEqual(501, response.code)



# -*- test-case-name: mimic.test.test_valkyrie -*-
"""
API Mock for Valkyrie.
"""

from mimic.rest.mimicapp import MimicApp


class ValkyrieApi(object):
    """
    Rest endpoints for the Valkyrie API.
    """

    app = MimicApp()

    def __init__(self, core):
        """
        :param MimicCore core: The core to which the Valkyrie Api will be
        communicating.
        """
        self.core = core

    @app.route('/login', methods=['POST'])
    def login(self, request):
        """
        Responds with response code 200 and returns an auth token
        See https://valkyrie.my.rackspace.com/#authentication
        """
        return self.core.valkyrie_store.create_token(request)

    @app.route('/login_user', methods=['POST'])
    def login_user(self, request):
        """
        Responds with response code 200 and returns an auth token
        See https://valkyrie.my.rackspace.com/#authentication
        """
        return self.core.valkyrie_store.create_token(request)

    effective_any_permissions_route = ('/account/<int:account_number>'
                                       '/permissions/contacts/any'
                                       '/by_contact/<int:contact_id>/effective')

    @app.route(effective_any_permissions_route, methods=['GET'])
    def effective_any_permissions(self, request, account_number, contact_id):
        """
        Responds with response code 200 and returns a list of all permissions
        for the given account and contact
        See https://valkyrie.my.rackspace.com/#managed-accounts
        """
        return self.core.valkyrie_store.get_permissions(request,
                                                        account_number, contact_id, None)

    effective_accounts_permissions_route = ('/account/<int:account_number>'
                                            '/permissions/contacts/accounts'
                                            '/by_contact/<int:contact_id>/effective')

    @app.route(effective_accounts_permissions_route, methods=['GET'])
    def effective_accounts_permissions(self, request, account_number, contact_id):
        """
        Responds with response code 200 and returns a list of account level permissions
        for the given account and contact
        See https://valkyrie.my.rackspace.com/#managed-accounts
        """
        return self.core.valkyrie_store.get_permissions(request,
                                                        account_number, contact_id, 1)

    effective_devices_permissions_route = ('/account/<int:account_number>'
                                           '/permissions/contacts/devices'
                                           '/by_contact/<int:contact_id>/effective')

    @app.route(effective_devices_permissions_route, methods=['GET'])
    def effective_devices_permissions(self, request, account_number, contact_id):
        """
        Responds with response code 200 and returns a list of device level permissions
        for the given account and contact
        See https://valkyrie.my.rackspace.com/#managed-accounts
        """
        return self.core.valkyrie_store.get_permissions(request,
                                                        account_number, contact_id, 2)



# -*- test-case-name: mimic.test.test_ironic -*-
"""
API Mock for Ironic.
http://docs.openstack.org/developer/ironic/webapi/v1.html
"""

from mimic.rest.mimicapp import MimicApp


class IronicApi(object):
    """
    Rest endpoints for the Ironic API.
    """

    app = MimicApp()

    def __init__(self, core):
        """
        :param MimicCore core: The core to which the Ironic Api will be
        communicating.
        """
        self.core = core

    @app.route('/nodes', methods=['POST'])
    def create_node(self, request):
        """
        Responds with response code 201 and returns the newly created node.
        """
        return self.core.ironic_node_store.create_node(request)

    @app.route('/nodes/<string:node_id>', methods=['DELETE'])
    def delete_node(self, request, node_id):
        """
        Responds with response code 204 and delete the node.
        """
        return self.core.ironic_node_store.delete_node(request, node_id)

    @app.route('/nodes', methods=['GET'])
    def list_nodes(self, request):
        """
        Responds with response code 200 with a list of nodes.
        """
        return self.core.ironic_node_store.list_nodes(include_details=False)

    @app.route('/nodes/detail', methods=['GET'])
    def list_nodes_with_details(self, request):
        """
        Responds with response code 200 with a list of nodes and its details.
        """
        return self.core.ironic_node_store.list_nodes(include_details=True)

    @app.route('/nodes/<string:node_id>', methods=['GET'])
    def get_node_details(self, request, node_id):
        """
        Responds with response code 200 with details of the nodes.
        """
        return self.core.ironic_node_store.get_node_details(request, node_id)

    @app.route('/nodes/<string:node_id>/states/provision', methods=['PUT'])
    def set_node_provision_state(self, request, node_id):
        """
        Responds with response code 202 and sets the provision state of
        the node.
        """
        return self.core.ironic_node_store.set_node_provision_state(
            request, node_id)

    @app.route('/nodes/<string:node_id>/vendor_passthru/<string:method>', methods=['POST'])
    def vendor_passthru_cache_image(self, request, node_id, method):
        """
        Responds with response code 202 and sets the :obj:`Node`'s cache_image_id
        and cache_status.
        Returns 400 if `node_id` does not exist or if the `method` is not `cache_image`
        """
        return self.core.ironic_node_store.cache_image_using_vendor_passthru(
            request, node_id, method)



# -*- test-case-name: mimic.test.test_mailgun -*-
"""
API Mock for Mail Gun.
https://documentation.mailgun.com/api-sending.html
"""

import json
import time
import urlparse

from mimic.rest.mimicapp import MimicApp
from mimic.util.helper import seconds_to_timestamp

count = 0


class MailGunApi(object):
    """
    Rest endpoints for mocked Mail Gun api.
    """

    app = MimicApp()

    def __init__(self, core):
        """
        :param MimicCore core: The core to which this Mail Gun Api will be
        communicating.
        """
        self.core = core

    @app.route('/messages', methods=['POST'])
    def send_messages(self, request):
        """
        Responds with response code 200 with a generic response body.

        If the `to` address is `bademail@example.com` results in error 500
        and if it is `failingemail@example.com` results in 400.
        """
        content = urlparse.parse_qs(request.content.read())
        to_address = content.get('to')
        headers = {}
        for key, value in content.items():
            if key.startswith("h:") or key.startswith("v:"):
                headers[key] = value

        if 'bademail@example.com' in to_address:
            request.setResponseCode(500)
            global count
            count += 1
            return b''

        if 'failingemail@example.com' in to_address:
            request.setResponseCode(400)
            return b''

        request.setResponseCode(200)
        message_id = "{0}@samples.mailgun.org".format(
            seconds_to_timestamp(time.time(),
                                 "%Y%m%d%H%M%S.%f"))
        self.core.message_store.add_to_message_store(
            message_id=message_id,
            to=content.get('to')[0], msg_from=content.get('from'),
            subject=content.get('subject')[0], body=content.get('html'),
            custom_headers=headers)
        return json.dumps({
            "message": "Queued. Thank you.",
            "id": message_id})

    @app.route('/messages', methods=['GET'])
    def get_messages(self, request):
        """
        Responds with a 200 and list of of messages POSTed
        through the ``/messages`` endpoint.
        """
        filter_by_to = request.args.get("to")
        request.setResponseCode(200)
        return json.dumps(self.core.message_store.list_messages(filter_by_to))

    @app.route('/messages/500s', methods=['GET'])
    def get_messages_500_count(self, request):
        """
        Responds with a 200 and the number of messages resulting in 500s
        i.e. when the `to` address is `bademail@example.com`.

        This API endpoint is not supported by Mailgun and is present for testing
        purposes.
        """
        request.setResponseCode(200)
        return json.dumps({"count": count})

    @app.route('/messages/headers', methods=['GET'])
    def get_message_headers(self, request):
        """
        Responds with a 200 and returns the headers recieved when the
        message with the given `to` address in the query, was created.

        This API endpoint is not supported by Mailgun and is present for testing
        purposes.
        """
        request.setResponseCode(200)
        to_addr = request.args.get("to")
        msg = self.core.message_store.filter_message_by_to_address(to_addr)
        return json.dumps({msg.to: msg.custom_headers})



"""
Defines the preset values in the mimic api.
"""

get_presets = {"loadbalancers": {"lb_building": "On create load balancer, keeps the load balancer in "
                                                "building state for given seconds",
                                 "lb_error_state": "Puts the LB in error state, and such an LB can only"
                                                   "be deleted",
                                 "lb_pending_update": "Changes the load balancer to PENDING-UPDATE"
                                                      "state for the given number of seconds, any action"
                                                      "other than delete is performed on the server",
                                 "lb_pending_delete": "Changes the load balancer to PENDING-DELETE"
                                                      "state for the given seconds, when deleted"},
               "servers": {"create_server_failure": "{\"message\": \"given message\","
                                                    "\"code\": given code}",
                           "delete_server_failure": "{\"code\": given code,"
                                                    "\"times\": returns given code that many times}",
                           "invalid_image_ref": ["INVALID-IMAGE-ID", "1111", "image_ends_with_Z"],
                           "invalid_flavor_ref": ["INVALID-FLAVOR-ID", "8888", "-4", "1"],
                           "server_error": "sets server state to error on create",
                           "server_building": "sets the server to be in building state for given time"
                                              " in seconds"},
               "identity": {
                   # On ``validate_token`` the tokens listed below
                   # result in 'monitoring-service-admin' impersonator role.
                   "maas_admin_roles": [
                       "this_is_an_impersonator_token",
                       "this_is_an_impersonator_token_also",
                       "impersonate_watson",
                       "impersonate_creator",
                       "this_is_an_impersonator_token_also_2",
                       "impersonate_foo_token"],
                   # On ``validate_token`` the tokens listed below
                   # result in 'racker' impersonator role.
                   "racker_token": ["this_is_a_racker_token"],
                   # Tenants with user observer role
                   "observer_role": ["09876"],
                   # Tenants with user creator role
                   "creator_role": ["09090"],
                   # Tenants with user admin role
                   "admin_role": ["9999"],
                   # Tenants with this token result in a 401 when validating the token
                   "token_fail_to_auth": ["never-cache-this-and-fail-to-auth"],
                   # Users presenting these tokens have contact IDs that correspond
                   # to presets in the Valkyrie plugin...
                   "non_dedicated_observer": ["OneTwo"],
                   "non_dedicated_admin": ["ThreeFour"],
                   "dedicated_full_device_permission_holder": ["HybridOneTwo"],
                   "dedicated_account_permission_holder": ["HybridThreeFour"],
                   "dedicated_limited_device_permission_holder": ["HybridFiveSix"],
                   "dedicated_other_account_observer": ["HybridSevenEight"],
                   "dedicated_other_account_admin": ["HybridNineZero"]}}



# -*- test-case-name: mimic.test.test_rackconnect_v3 -*-

"""
API mock for the Rackspace RackConnect v3 API, which is documented at:
http://docs.rcv3.apiary.io/
"""
from collections import defaultdict
import json
from uuid import uuid4, UUID

from characteristic import attributes, Attribute
from six import text_type

from twisted.plugin import IPlugin
from twisted.web.http import NOT_FOUND, NOT_IMPLEMENTED
from zope.interface import implementer

from mimic.catalog import Entry
from mimic.catalog import Endpoint
from mimic.imimic import IAPIMock
from mimic.rest.mimicapp import MimicApp
from mimic.util.helper import random_ipv4, seconds_to_timestamp


timestamp_format = '%Y-%m-%dT%H:%M:%SZ'


@implementer(IAPIMock, IPlugin)
class RackConnectV3(object):
    """
    API mock object for RackConnect V3.

    :ivar regions: The regions for which endpoints should be produced
    :type regions: ``iterable``

    :ivar int default_pools: The number of default load balancers to be
        created per tenant per region
    """

    def __init__(self, regions=("ORD",), default_pools=1):
        """
        Construct a :class:`RackConnectV3` object.
        """
        self.regions = regions
        self.default_pools = default_pools

    def catalog_entries(self, tenant_id):
        """
        Catalog entry for RackConnect V3 endpoints.
        """
        return [
            Entry(tenant_id, "rax:rackconnect", "rackconnect", [
                Endpoint(tenant_id, region, text_type(uuid4()), prefix="v3")
                for region in self.regions
            ])
        ]

    def resource_for_region(self, region, uri_prefix, session_store):
        """
        Return an IResource implementing a public RackConnect V3 region
        endpoint.
        """
        return RackConnectV3Region(
            iapi=self,
            uri_prefix=uri_prefix,
            session_store=session_store,
            region_name=region,
            default_pools=self.default_pools).app.resource()


@attributes(
    [Attribute("id", default_factory=lambda: text_type(uuid4()),
               instance_of=text_type),
     Attribute("name", default_value=u"default", instance_of=text_type),
     Attribute("port", default_value=80, instance_of=int),
     Attribute("status", default_value=u"ACTIVE", instance_of=text_type),
     Attribute("status_detail", default_value=None),
     Attribute("virtual_ip", default_factory=random_ipv4,
               instance_of=text_type),
     Attribute('nodes', default_factory=list, instance_of=list)],
    apply_with_cmp=False)
class LoadBalancerPool(object):
    """
    Represents a RackConnecvt v3 Load Balancer Pool.

    Load balancer pools cannot be created via the API, they have to just
    already exist on the account.

    :param text_type id: The ID of the load balancer pool - is randomly
        generated if not provided
    :param text_type name: The name of the load balancer pool - defaults to
        'default'
    :param int port: The incoming port of the load balancer - defaults to 80
    :param text_type status: The status of the load balancer - defaults to
        "ACTIVE"
    :param text_type status_detail: Any details about the status - defaults
        to None
    :param text_type virtual_ip: The IP of the load balancer pool
    :param list nodes: :class:`LoadBalancerPoolNode`s
    """
    def as_json(self):
        """
        Create a JSON-serializable representation of the contents of this
        object, which can be used in a REST response for a request for the
        details of this particular object
        """
        # no dictionary comprehensions in py2.6
        response = dict([
            (attr.name, getattr(self, attr.name))
            for attr in LoadBalancerPool.characteristic_attributes
            if attr.name != "nodes"])
        response['node_counts'] = {
            "cloud_servers": len(self.nodes),
            "external": 0,
            "total": len(self.nodes)
        }
        return response

    def node_by_cloud_server(self, cloud_server_id):
        """
        Find a node by its cloud server ID.

        :return: a :class:`LoadBalancerPoolNode` with a server id matching
            the given cloud server, or `None` if none are found
        """
        return next((node for node in self.nodes
                     if node.cloud_server == cloud_server_id),
                    None)

    def node_by_id(self, node_id):
        """
        Find a node by its node_id.

        :return: a :class:`LoadBalancerPoolNode` with the given node id,
            or `None` if none are found
        """
        return next((node for node in self.nodes if node.id == node_id), None)


@attributes(["created", "load_balancer_pool", "cloud_server",
             Attribute("id", default_factory=lambda: text_type(uuid4()),
                       instance_of=text_type),
             Attribute("updated", default_value=None),
             Attribute("status", default_value=text_type("ACTIVE"),
                       instance_of=text_type),
             Attribute("status_detail", default_value=None)])
class LoadBalancerPoolNode(object):
    """
    Represents a Load Balancer Pool Node.

    A load currently only corresponds to a particular cloud server.
    External nodes are not currently supported in RackConnectV3

    :param text_type id: The ID of the load balancer pool node - is randomly
        generated if not provided
    :param text_type created: The timestamp of when the load balancer node was
        created
    :param text_type updated: The timestamp of when the load balancer node was
        updated.  Can be None.
    :param load_balancer_pool: :class:`LoadBalancerPool` to which
        this node belongs
    :param text_type status: The status of the load balancer - defaults to
        "ACTIVE"
    :param text_type status_detail: Any details about the status - defaults
        to None
    :param text_type cloud_server: The ID of the cloud server corresponding to
        this node.  The cloud server's ID is not equivalent to the ``id``
        attribute on :class:`LoadBalancerPoolNode`, because a node
        in theory can also be some external (not a cloud server) resource,
        but that is not supported yet on the API.
    """
    def short_json(self):
        """
        Create a short JSON-serializable representation of the contents of
        this object (not detailed), which can be used in a REST response.

        Should match the JSON documented in the response for
        POST /v3/{tenant_id}/load_balancer_pools/{load_balancer_pool_id}/nodes
        (add load balancer pool node) or as part of the response for
        GET /v3/{tenant_id}/load_balancer_pools/{load_balancer_pool_id}/nodes
        (list load balancer pool nodes)
        """
        # no dictionary comprehensions in py2.6
        response = dict([
            (attr.name, getattr(self, attr.name))
            for attr in LoadBalancerPoolNode.characteristic_attributes
            if attr.name not in ('load_balancer_pool', 'cloud_server')])
        response['load_balancer_pool'] = {'id': self.load_balancer_pool.id}
        response['cloud_server'] = {'id': self.cloud_server}
        return response

    def update(self, now, status, status_detail=None):
        """
        Changes the status of the node.
        """
        self.updated = now
        self.status = status
        self.status_detail = status_detail


@attributes(["iapi", "uri_prefix", "session_store", "region_name",
             "default_pools"])
class RackConnectV3Region(object):
    """
    A set of ``klein`` routes representing a RackConnect V3 endpoint.
    """
    app = MimicApp()

    @app.route("/v3/<string:tenant_id>/load_balancer_pools", branch=True)
    def get_tenant_lb_pools(self, request, tenant_id):
        """
        Get a resource for a tenant's load balancer pools in this region.
        """
        tenant_store = self.session_store.session_for_tenant_id(tenant_id)
        per_tenant_lbs = tenant_store.data_for_api(
            self.iapi, lambda: defaultdict(list))
        per_tenant_per_region_lbs = per_tenant_lbs[self.region_name]

        if not per_tenant_per_region_lbs:
            per_tenant_per_region_lbs.extend([
                LoadBalancerPool() for _ in range(self.default_pools)])

        handler = LoadBalancerPoolsInRegion(lbpools=per_tenant_per_region_lbs,
                                            clock=self.session_store.clock)
        return handler.app.resource()


# exclude all the attributes from comparison so that equality has to be
# determined by identity, since lbpools is mutable and we don't want to
# compare clocks
@attributes(["lbpools", "clock"], apply_with_cmp=False)
class LoadBalancerPoolsInRegion(object):
    """
    A set of ``klein`` routes handling RackConnect V3 Load Balancer Pools
    collections.
    """
    app = MimicApp()

    def _pool_by_id(self, id):
        """
        Get a pool by the ID of the pool.
        """
        return next((pool for pool in self.lbpools if pool.id == id), None)

    @app.route("/", methods=["GET"])
    def list_all_load_balancer_pools(self, request):
        """
        API call to list all load balancer pools for the tenant and region
        correspoding to this handler.  Returns 200 always.

        http://docs.rcv3.apiary.io/#get-%2Fv3%2F%7Btenant_id%7D%2Fload_balancer_pools
        """
        return json.dumps([pool.as_json() for pool in self.lbpools])

    @app.route("/nodes", methods=["POST"])
    def bulk_add_nodes_to_load_balancer_pools(self, request):
        """
        Add multiple nodes to multiple load balancer pools.

        Returns a 400 if the lb pool_id is not a uuid.
        Returns a 409 if the lb pool_id does not exist or the cloud server already exists
        on the lb pool.

        http://docs.rcv3.apiary.io/#post-%2Fv3%2F%7Btenant_id%7D%2Fload_balancer_pools%2Fnodes

        TODO: blow up with a 500 and verify if the given server exists in nova.
        """
        body = json.loads(request.content.read())
        added_nodes = []
        error_response = {"errors": []}

        for each in body:
            pool_id = each['load_balancer_pool']['id']
            try:
                UUID(pool_id, version=4)
            except (ValueError, AttributeError):
                request.setResponseCode(400)
                return json.dumps('The input was not in the correct format. Please reference '
                                  'the documentation at http://docs.rcv3.apiary.io for '
                                  'further assistance.')
            pool = self._pool_by_id(pool_id)
            if pool is None:
                response_code = 409
                error_response["errors"].append("Load Balancer Pool {0} does "
                                                "not exist".format(pool_id))
            elif pool.node_by_cloud_server(each['cloud_server']['id']):
                response_code = 409
                error_response["errors"].append("Cloud Server {0} is already a "
                                                "member of Load Balancer Pool "
                                                "{1}".format(each['cloud_server']['id'],
                                                             pool_id))
        if not error_response['errors']:
            for add in body:
                node = LoadBalancerPoolNode(
                    created=seconds_to_timestamp(self.clock.seconds(),
                                                 timestamp_format),
                    load_balancer_pool=pool,
                    cloud_server=add['cloud_server']['id'])

                pool.nodes.append(node)
                added_nodes.append(node)
                response_code = 201

        request.setResponseCode(response_code)
        if response_code == 201:
            return json.dumps([n.short_json() for n in added_nodes])
        else:
            return json.dumps(error_response)

    @app.route("/nodes", methods=["DELETE"])
    def bulk_delete_nodes_to_load_balancer_pools(self, request):
        """
        Delete multiple nodes from multiple load balancer pools.

        Returns a 400 if the lb pool_id is not a uuid.
        Returns a 409 if the lb pool_id does not exist or the cloud server does not exist
        on the lb pool.

        http://docs.rcv3.apiary.io/#delete-%2Fv3%2F%7Btenant_id%7D%2Fload_balancer_pools%2Fnodes

        TODO: For now, blow up with a 500 and verify if the given server exists in nova.
        """
        body = json.loads(request.content.read())
        error_response = {"errors": []}

        for each in body:
            pool_id = each['load_balancer_pool']['id']
            try:
                UUID(pool_id, version=4)
            except (ValueError, AttributeError):
                request.setResponseCode(400)
                return json.dumps('The input was not in the correct format. Please reference '
                                  'the documentation at http://docs.rcv3.apiary.io for '
                                  'further assistance.')
            pool = self._pool_by_id(pool_id)

            if pool is None:
                response_code = 409
                error_response["errors"].append("Load Balancer Pool {0} does "
                                                "not exist".format(pool_id))
            elif pool.node_by_cloud_server(each['cloud_server']['id']) is None:
                response_code = 409
                error_response["errors"].append(("Cloud Server {0} is not a "
                                                 "member of Load Balancer Pool "
                                                 "{1}").format(each['cloud_server']['id'],
                                                               pool_id))
        if not error_response['errors']:
            for add in body:
                pool = self._pool_by_id(add['load_balancer_pool']['id'])
                node = pool.node_by_cloud_server(add['cloud_server']['id'])
                pool.nodes.remove(node)
                response_code = 204

        request.setResponseCode(response_code)
        if response_code == 204:
            return b""
        else:
            return json.dumps(error_response)

    @app.route("/<string:id>", branch=True)
    def delegate_to_one_pool_handler(self, request, id):
        """
        If the load balancer pool of the given ID exists, delgate to the
        :class:`OneLoadBalancerPool` handler for further requests.

        Returns 400 is the pool_id is not a uuid
        Returns 404 if no pool with the ID exists.
        """
        try:
            UUID(id, version=4)
        except (ValueError, AttributeError):
            request.setResponseCode(400)
            return ('The input was not in the correct format. Please reference '
                    'the documentation at http://docs.rcv3.apiary.io for further assistance.')
        pool = self._pool_by_id(id)
        if pool is not None:
            handler = OneLoadBalancerPool(pool=pool)
            app = handler.app
            resource = app.resource()
            return resource

        # TODO: what is the error message if any?  It is undocumented.
        # guess based on Remove Load Balancer Pool Node documentation
        request.setResponseCode(NOT_FOUND)
        return "Load Balancer Pool {0} does not exist".format(id)


@attributes(["pool"])
class OneLoadBalancerPool(object):
    """
    A set of ``klein`` routes handling the RackConnect V3 API for a single
    load balancer pool
    """
    app = MimicApp()

    @app.route("/", methods=["GET"])
    def get_pool_information(self, request):
        """
        API call to get the details of the single load balancer pool
        corresponding to this handler.

        Returns a 200 because the pool definitely exists by the time this
        handler is invoked.

        http://docs.rcv3.apiary.io/#get-%2Fv3%2F%7Btenant_id%7D%2Fload_balancer_pools%2F%7Bid%7D
        """
        return json.dumps(self.pool.as_json())

    @app.route("/nodes", methods=["GET"])
    def get_node_collection_information(self, request):
        """
        List all the nodes for the load balancer pool.

        Returns a 200 always.

        http://docs.rcv3.apiary.io/#get-%2Fv3%2F%7Btenant_id%7D%2Fload_balancer_pools%2F%7Bload_balancer_pool_id%7D%2Fnodes
        """
        return json.dumps([node.short_json() for node in self.pool.nodes])

    @app.route("/nodes/details", methods=["GET"])
    def get_node_collection_details_information(self, request):
        """
        Get detailed information about all the nodes, including the cloud
        server network information.  This is unimplemented as it might
        need to be hooked up to the nova api.

        http://docs.rcv3.apiary.io/#get-%2Fv3%2F%7Btenant_id%7D%2Fload_balancer_pools%2F%7Bload_balancer_pool_id%7D%2Fnodes%2Fdetails
        """
        request.setResponseCode(NOT_IMPLEMENTED)

    @app.route("/nodes", methods=["POST"])
    def add_single_pool_node(self, request):
        """
        Add a single pool node to the load balancer pool is not implemented
        yet.

        http://docs.rcv3.apiary.io/#post-%2Fv3%2F%7Btenant_id%7D%2Fload_balancer_pools%2F%7Bload_balancer_pool_id%7D%2Fnodes
        """
        request.setResponseCode(NOT_IMPLEMENTED)

    @app.route("/nodes/<string:node_id>", branch=True)
    def handle_single_node_requests(self, request, node_id):
        """
        Catchall to specify that single node operations (GET, DELETE,
        GET details) are not implemented yet
        """
        request.setResponseCode(NOT_IMPLEMENTED)
